import unittest

from ..config import parse_environment


class ParseEnvironmentTestCase(unittest.TestCase):

    def test_unsupported_engine_raises_valueerror(self):
        env = {
            'DB_ENGINE': 'foo',
        }
        with self.assertRaises(ValueError):
            opts = parse_environment(env)

    def test_empty_environment_returns_none(self):
        opts = parse_environment({})
        self.assertEqual(opts, None)

    def test_parse_postgresql_with_defaults(self):
        env = {
            'DB_ENGINE': 'postgresql',
        }
        opts = parse_environment(env)
        self.assertEqual(opts.host, 'localhost')
        self.assertEqual(opts.port, 5432)
        self.assertEqual(opts.name, None)
        self.assertEqual(opts.username, None)
        self.assertEqual(opts.password, None)

    def test_parse_mysql_with_defaults(self):
        env = {
            'DB_ENGINE': 'mysql',
        }
        opts = parse_environment(env)
        self.assertEqual(opts.host, 'localhost')
        self.assertEqual(opts.port, 3306)
        self.assertEqual(opts.name, None)
        self.assertEqual(opts.username, None)
        self.assertEqual(opts.password, None)

    def test_parse_sqlite_with_defaults(self):
        env = {
            'DB_ENGINE': 'sqlite',
        }
        opts = parse_environment(env)
        self.assertEqual(opts.host, None)
        self.assertEqual(opts.port, None)
        self.assertEqual(opts.name, ":memory:")
        self.assertEqual(opts.username, None)
        self.assertEqual(opts.password, None)



import copy
import os

from unimatrix.const import SECDIR
from unimatrix.lib.datastructures import DTO


__all__ = ['parse_environment']


def _parse_bool(value):
    if value is None:
        return None
    return value == '1'


ENVIRONMENT_VARIABLES = [
    ('DB_HOST', 'host', False, str),
    ('DB_PORT', 'port', False, int),
    ('DB_NAME', 'name', False, str),
    ('DB_USERNAME', 'username', False, str),
    ('DB_PASSWORD', 'password', False, str),
    ('DB_CONNECTION_MAX_AGE', 'max_age', False, int),
    ('DB_AUTOCOMMIT', 'autocommit', False, _parse_bool)
]


#: Default values by engine.
DEFAULTS = {
    'postgresql': {
        'host': "localhost",
        'port': 5432,
    },
    'mysql': {
        'host': "localhost",
        'port': 3306,
    },
    'sqlite': {
        'name': ":memory:"
    }
}


def _get_computed_value(engine, name, opts):
    computed = None
    if name == 'host':
        computed = DEFAULT_HOSTS[engine]
    elif name == 'port':
        computed = DEFAULT_PORTS[engine]   
    else:
        raise NotImplementedError
    return computed


def parse_environment(env):
    """Parses a database connection from the operating system
    environment based on the ``DB_*`` variables. Return a
    datastructure containing the configuration, or ``None`` if
    it was not defined.
    """
    # If no engine is specified, then the application does not
    # get its database connection configuration from environment
    # variables.
    engine = env.get('DB_ENGINE')
    if engine is None:
        return None

    if engine not in DEFAULTS:
        raise ValueError("Unsupported DB_ENGINE: %s" % engine)

    opts = DTO()
    for name, attname, required, cls in ENVIRONMENT_VARIABLES:
        default = DEFAULTS[engine].get(attname)
        value = env.get(name)
        if value is None:
            value = default
        opts[attname] = cls(value)
        if required and not value and not default:
            raise NotImplementedError

    opts.engine = engine
    return opts


def load(env=None, config_dir=os.path.join(SECDIR, 'rdbms/connections')):
    """Loads the configured database connections from the operating
    system environment variables and the Unimatrix configuration files.

    Database connections are loaded with the following precedence:

    - Environment variables
    - Unimatrix Database Connection (UDC) files.

    Args:
        env (dict): a dictionary holding environment variables. Defaults
            to ``os.environ``.
        config_dir (str): a directory on the local filesystem holding
            the database connection configuration files. Defaults to
            `SECDIR/rdbms/connections`.

    Returns:
        dict
    """
    connections = {}
    self = parse_environment(copy.deepcopy(env or os.environ))
    if self:
        connections['self'] = self
    return connections



import functools


class FailedOperation(Exception):
    """Represents an exception that was caused by a failure
    of a requested operation.
    """
    default_status = 421
    default_code = 'UNKNOWN_ERROR'

    @classmethod
    def catch(cls, on_exception):
        """Returns a decorator that invokes callable `on_exception`
        when a :class:`FailedOperation` is raised.
        """
        def decorator(func):
            @functools.wraps(func)
            def f(*args, **kwargs):
                try:
                    return func(*args, **kwargs)
                except FailedOperation as e:
                    return on_exception(e)
            return f
        return decorator

    def __init__(self, status=None, code=None, stderr=None):
        self.code = code or self.default_code
        self.status = status or self.default_status
        self.stderr = stderr or ''

    @property
    def dto(self):
        return {
            'code': self.code,
            'stderr': self.stderr
        }


class AccessDenied(FailedOperation):
    """Raised when an operation fails due to request not
    having permission.
    """
    default_code = 'ACCESS_DENIED'
    default_status = 401


class ResourceDoesNotExist(FailedOperation):
    """Raised when an operation fails due to an upstream
    resource not existing.
    """
    default_code = 'DOES_NOT_EXIST'
    default_status = 404


class Unreachable(FailedOperation):
    """Raised when an operation fails due to an upstream
    resource not being reachable.
    """
    default_code = 'UNREACHABLE'
    default_status = 503


class Timeout(Unreachable):
    """Raised when an operation fails due to a timeout
    when accessing an upstream resource.
    """
    pass





DJANGO_DOMAIN               = 'django.python.unimatrixone.io'
DJANGO_PROJECT_NAME         = f'{DJANGO_DOMAIN}/project'
DJANGO_SETTINGS_MODULE      = f'{DJANGO_DOMAIN}/settings'
PLATFORM_DOMAIN             = "platform.unimatrixone.io"
PLATFORM_LANGUAGE_NAME      = f"{PLATFORM_DOMAIN}/language-name"
PLATFORM_LANGUAGE_VERSION   = f"{PLATFORM_DOMAIN}/language-version"
PLATFORM_FRAMEWORK_NAME     = f"{PLATFORM_DOMAIN}/framework-name"
PLATFORM_FRAMEWORK_VERSION  = f"{PLATFORM_DOMAIN}/framework-version"
PLATFORM_OS_NAME            = f"{PLATFORM_DOMAIN}/os-name"
PLATFORM_OS_RELEASE         = f"{PLATFORM_DOMAIN}/os-release"
